/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include <cutil_inline.h>

#include "dAdq_cuda.h"

#define BLOCK_SIZE 8
#define BLOCKS(s) ((s)/BLOCK_SIZE + (((s)%BLOCK_SIZE) ? 1 : 0))

// all arrays column-minor
// Ns a multiple of BLOCK_SIZE, arrays padded with 0 as needed.

// gklt[k<Ns][l<Ns]
// dQdq[q<Nq][i<Ns][j<Ns]
// Aek [e<Nstim][k<Ns][i<Ns][j<Ns]
// dAdq_kl[q<Nq][k<Ns][l<Ns][i<Ns][j<Ns]
// dAdq[q<Nq][i<Ns][j<Ns]



// grid size: (Nq * Nk * Nk) x (BLOCKS(Ns) * BLOCKS(Ns))
__global__ void dAdq_zero_kernel(int Nstim, int Ns, int Nk, int Nq, float *gklt, float *dQdq, float *Aek, float *dAdq_kl, float *dAdq)
{
  int l = blockIdx.x % Nk;
  int i = blockIdx.x / Nk;
  int k = i % Nk;
  int q = i / Nk;
  i = BLOCK_SIZE * (blockIdx.y / BLOCKS(Ns)) + threadIdx.x;
  int j = BLOCK_SIZE * (blockIdx.y % BLOCKS(Ns)) + threadIdx.y;
  
  int stride_l = Ns*Ns;
  int stride_k = stride_l*Nk;
  int stride_q = stride_k*Nk;

  dAdq_kl[q*stride_q + k*stride_k + l*stride_l + Ns*i + j] = 0.0f;
  if ( (q|l) == 0 )
    for (int e=0; e<Nstim; ++e)
      Aek[e*stride_k + k*stride_l + i*Ns + j] = 0.0f;
  if ( (k|l) == 0 ) {
    dQdq[q*stride_l + i*Ns + j] = 0.0f;
    dAdq[q*stride_l + i*Ns + j] = 0.0f;
  }
  if ( (q|k|l) == 0 )
    gklt[i*Ns + j] = 0.0f;
}

// grid size: (Nq * Nk * Nk) x (BLOCKS(Ns) * BLOCKS(Ns))  
__global__ void dAdq_kl_kernel(int Ns, int Nk, int Nq, float *gklt, float *dQdq, float *Ak, float *dAdq_kl)
{
  int l = blockIdx.x % Nk;
  int i = blockIdx.x / Nk;
  int k = i % Nk;
  int q = i / Nk;
  i = BLOCK_SIZE * (blockIdx.y / BLOCKS(Ns)) + threadIdx.x;
  int j = BLOCK_SIZE * (blockIdx.y % BLOCKS(Ns)) + threadIdx.y;
  
  __shared__ float A[BLOCK_SIZE][BLOCK_SIZE];
  __shared__ float B[BLOCK_SIZE][BLOCK_SIZE];
  __shared__ float C[BLOCK_SIZE][BLOCK_SIZE];
  float tmp = 0.0f;
  
  for (int mb=0; mb<BLOCKS(Ns); ++mb) {
    A[threadIdx.x][threadIdx.y] = Ak[l*Ns*Ns + (mb*BLOCK_SIZE+threadIdx.x)*Ns + j];
    
    for (int m=0; (m<BLOCK_SIZE)&&((mb*BLOCK_SIZE+m)<Nk); ++m) {
      float tmp_inner = 0.0f;
      
      for (int nb=0; nb<BLOCKS(Ns); ++nb) {
	B[threadIdx.x][threadIdx.y] = dQdq[q*Ns*Ns + (nb*BLOCK_SIZE+threadIdx.x)*Ns + (mb*BLOCK_SIZE+threadIdx.y)];
	C[threadIdx.x][threadIdx.y] = Ak[k*Ns*Ns + i*Ns + (nb*BLOCK_SIZE+threadIdx.y)];
	__syncthreads(); // B[n] may be from a different warp; this also syncs A[m]
	
	for (int n=0; (n<BLOCK_SIZE)&&((nb*BLOCK_SIZE+n)<Nk); ++n)
	  tmp_inner += B[n][m] * C[threadIdx.x][n];
      }
      tmp += A[m][threadIdx.y] * tmp_inner;
    }
  }
  
  int stride_l = Ns*Ns;
  int stride_k = stride_l*Nk;
  int stride_q = stride_k*Nk;
  dAdq_kl[q*stride_q + k*stride_k + l*stride_l + i*Ns + j] = gklt[k*Ns + l] * tmp;
}


// grid size: Nq x (BLOCKS(Ns) * BLOCKS(Ns))
__global__ void dAdq_sum_kernel(int Ns, int Nk, int Nq, float *dAdq_kl, float *dAdq)
{
  // q: blockIdx.x
  int i = BLOCK_SIZE * (blockIdx.y / BLOCKS(Ns)) + threadIdx.x;
  int j = BLOCK_SIZE * (blockIdx.y % BLOCKS(Ns)) + threadIdx.y;
  int stride_l = Ns*Ns;
  int stride_k = stride_l*Nk;
  int stride_q = stride_k*Nk;
  float tmp = 0.0;
  for (int k=0; k<Nk; ++k)
    for (int l=0; l<Nk; ++l)
      tmp += dAdq_kl[blockIdx.x*stride_q + k*stride_k + l*stride_l + i*Ns + j];
  dAdq[blockIdx.x*Ns*Ns + i*Ns + j] = tmp;
}

extern "C" void dAdq_zero_kernel_call(int Nstim, int Ns, int Nk, int Nq, float *gklt, float *dQdq, float *Aek, float *dAdq_kl, float *dAdq)
{
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(Nq*Nk*Nk, BLOCKS(Ns)*BLOCKS(Ns));
  dAdq_zero_kernel<<<dimGrid, dimBlock>>>(Nstim, Ns, Nk, Nq, gklt, dQdq, Aek, dAdq_kl, dAdq);
}

extern "C" void dAdq_kl_kernel_call(int Ns, int Nk, int Nq, float *gklt, float *dQdq, float *Ak, float *dAdq_kl)
{
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(Nq*Nk*Nk, BLOCKS(Ns)*BLOCKS(Ns));
  dAdq_kl_kernel<<<dimGrid, dimBlock>>>(Ns, Nk, Nq, gklt, dQdq, Ak, dAdq_kl);
}

extern "C" void dAdq_sum_kernel_call(int Ns, int Nk, int Nq, float *dAdq_kl, float *dAdq)
{
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(Nq, BLOCKS(Ns)*BLOCKS(Ns));
  dAdq_sum_kernel<<<dimGrid, dimBlock>>>(Ns, Nk, Nq, dAdq_kl, dAdq);
}



